@@ -3,7 +3,7 @@ package Params::Sah;
 our $AUTHORITY = 'cpan:PERLANCAR'; # AUTHORITY
 our $DATE = '2020-05-08'; # DATE
 our $DIST = 'Params-Sah'; # DIST
-our $VERSION = '0.070'; # VERSION
+our $VERSION = '0.071'; # VERSION
 
 use 5.010001;
 use strict;
@@ -13,16 +13,16 @@ use Log::ger;
 use Exporter qw(import);
 our @EXPORT_OK = qw(gen_validator);
 
-our $OPT_ON_INVALID  = 'croak';
-our $OPT_NAMED       = 0;
-our $OPT_DISABLE     = 0;
-our $OPT_ALLOW_EXTRA = 0;
-
+our $OPT_ON_INVALID     = 'croak';
+our $OPT_INVALID_DETAIL = 0;
+our $OPT_NAMED          = 0;
+our $OPT_DISABLE        = 0;
+our $OPT_ALLOW_EXTRA    = 0;
 our $DEBUG;
 
 sub gen_validator {
     my ($opt_on_invalid, $opt_named, $opt_disable, $opt_allow_extra,
-        $opt_optional_params);
+        $opt_invalid_detail, $opt_optional_params);
     {
         my $opts;
         if (ref($_[0]) eq 'HASH') {
@@ -32,9 +32,10 @@ sub gen_validator {
         }
         $opt_on_invalid = delete $opts->{on_invalid} // $OPT_ON_INVALID //
             'croak';
-        die "Invalid on_invalid value, must be: croak|carp|warn|die|bool|str"
-            unless $opt_on_invalid =~ /\A(croak|carp|warn|die|bool|str)\z/;
+        die "Invalid on_invalid value, must be: croak|carp|warn|die|return"
+            unless $opt_on_invalid =~ /\A(croak|carp|warn|die|return)\z/;
 
+        $opt_invalid_detail = delete $opts->{invalid_detail} // $OPT_INVALID_DETAIL // 0;
         $opt_named = delete $opts->{named} // $OPT_NAMED // 0;
         $opt_disable = delete $opts->{disable} // $OPT_DISABLE // 0;
         $opt_allow_extra = delete $opts->{allow_extra} // $OPT_ALLOW_EXTRA // 0;
@@ -75,16 +76,17 @@ sub gen_validator {
     my %mentioned_vars;
 
     my $code_get_err_stmt = sub {
-        my $arg_term = shift;
+        my ($err_term_detail, $err_term_generic) = @_;
         if ($opt_on_invalid =~ /\A(croak|carp|warn|die)\z/) {
             my $stmt = $opt_on_invalid =~ /\A(croak|carp)\z/ ?
                 "Carp::$opt_on_invalid" : $opt_on_invalid;
-            return "$stmt($arg_term)";
+            return "$stmt(".($opt_invalid_detail ? $err_term_detail : $err_term_generic // $err_term_detail).")";
         } else {
-            if ($opt_on_invalid eq 'bool') {
-                return "return 0";
+            # return
+            if ($opt_invalid_detail) {
+                return "return $err_term_detail";
             } else {
-                return "return $arg_term";
+                return "return 0";
             }
         }
     };
@@ -98,7 +100,7 @@ sub gen_validator {
     $src .= "sub {\n";
 
     $src .= "    my \$_ps_args = shift;\n";
-    $src .= "    my \$_ps_res;\n" unless $opt_on_invalid eq 'bool';
+    $src .= "    my \$_ps_res;\n" if $opt_invalid_detail;
 
     unless ($opt_allow_extra) {
         $src .= "\n    ### checking unknown arguments\n";
@@ -138,7 +140,7 @@ sub gen_validator {
             data_term    => $argterm,
             err_term     => '$_ps_res',
             schema       => $schemas{$argname},
-            return_type  => $opt_on_invalid eq 'bool' ? 'bool' : 'str',
+            return_type  => $opt_invalid_detail ? 'str' : 'bool',
             indent_level => 1,
         );
         die "Incompatible Data::Sah version (cd v=$cd->{v}, expected 2)" unless $cd->{v} == 2;
@@ -157,14 +159,14 @@ sub gen_validator {
         }
         $src .= "    undef \$_ps_res;\n" if
             $i && $opt_on_invalid =~ /\A(carp|warn)\z/;
-        $src .= "    ".$code_get_err_stmt->(qq("$data_name: \$_ps_res"))." if !($cd->{result});\n";
+        $src .= "    ".$code_get_err_stmt->(qq("$data_name: \$_ps_res"), qq("$data_name: fail schema ".).Data::Dmp::dmp($schemas{$argname}))." if !($cd->{result});\n";
         $i++;
     } # for $argname
 
-    if ($opt_on_invalid eq 'bool') {
-        $src .= "\n    return 1\n";
-    } else {
+    if ($opt_invalid_detail) {
         $src .= "\n    return '';\n";
+    } else {
+        $src .= "\n    return 1;\n";
     }
 
     $src .= "\n};";
@@ -194,7 +196,7 @@ Params::Sah - Validate method/function parameters using Sah schemas
 
 =head1 VERSION
 
-This document describes version 0.070 of Params::Sah (from Perl distribution Params-Sah), released on 2020-05-08.
+This document describes version 0.071 of Params::Sah (from Perl distribution Params-Sah), released on 2020-05-08.
 
 =head1 SYNOPSIS
 
@@ -302,6 +304,10 @@ Bool. Used to set default for C<allow_extra> option.
 
 String. Used to set default for C<on_invalid> option.
 
+=head2 $OPT_ERR_DETAIL
+
+Bool. Used to set default for C<err_detail> option.
+
 =head2 $OPT_DISABLE
 
 Bool. Used to set default for C<disable> option.
@@ -363,6 +369,18 @@ perspective. Other valid choices include: C<warn>, C<carp>, C<die>, C<bool>
 (return false on invalid, or true on valid), C<str> (return an error message on
 invalid, or empty string on valid).
 
+=item * invalid_detail => bool (default: 0)
+
+If set to true, will generate a more detailed error message. For example, with
+this schema:
+
+ [str => {min_len=>4}]
+
+then the string C<'foo'> will fail to validate with this error message "Length
+must be at least 4". Otherwise, the error message will just be something like:
+"Fail schema ['str', {min_len=>1}]". By default this option is set to false for
+slightly faster validation.
+
 =item * disable => bool (default: 0)
 
 If set to 1, will return an empty coderef validator. Used to disable parameter
